﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SimplerHid;

namespace Mcp2210
{
    public partial class Mcp2210Device
    {
        #region Variables
        private HidDevice device = null;
        public Mcp2210Gpio[] Gpios = new Mcp2210Gpio[9];
        internal SpiTransferSettings spiSettings = new SpiTransferSettings();
        internal GpioSettings gpioSettings = new GpioSettings();
        public bool Immediate = true;
        #endregion

        /// <summary>
        /// Sets the way interrupts are counted for GPIO6.
        /// When Immediate of Mcp2210Device is false, [GS]etGpioChipSettings must be called to retreive/apply the value.
        /// </summary>
        public InterruptMode Gpio6IrqMode
        {
            get
            {
                if(Immediate == true)
                {
                    GetGpioChipSettings();
                }

                int tmp = (gpioSettings.otherChipSettings >> 1) & 0x07;
                if (tmp > 4)
                {
                    throw new Exception("Value for IRQ mode is out of range");
                }
                return (InterruptMode)tmp;
            }
            set
            {
                int tmp = gpioSettings.otherChipSettings & 0x0E;
                tmp |= ((int)value) << 1;
                gpioSettings.otherChipSettings = (byte)tmp;

                if(Immediate == true)
                {
                    SetGpioChipSettings();
                }
            }
        }

        public Mcp2210Device(string devicePath)
        {
            device = new HidDevice(devicePath);
            for (int i = 0; i < Gpios.Length; i++)
            {
                this.Gpios[i] = new Mcp2210Gpio(this, (byte)i);
            }

            getSpiSettings();
            GetGpioChipSettings();
        }

        public void Dispose()
        {
            device.Dispose();
        }

        private object sendCommand(object cmd, Type respType)
        {
            byte[] tmp = HidDevice.GetBytesFromObject(cmd);

            device.WriteOutputReport(0, tmp);
            byte[] resp = device.ReadInputReport();

            if (resp[0] != tmp[0])
                throw new Exception("Response command code doesn't match request, no exclusive access?");

            return HidDevice.GetObjectFromBytes(resp, respType);
        }

        public ChipStatusResp GetChipStatus()
        {
            ChipStatusCmd cmd = new ChipStatusCmd()
            {
                cmdCode = Commands.GetChipStatus,
            };

            ChipStatusResp resp = (ChipStatusResp)sendCommand(cmd, typeof(ChipStatusResp));
            return resp;
        }


        public ushort GetInterrupts(bool reset = true)
        {
            EventCounterCmd cmd = new EventCounterCmd()
            {
                cmdCode = Commands.GetEventsCount,
                noReset = (byte)(reset ? 0 : 1),
            };

            EventCounterResp resp = (EventCounterResp)sendCommand(cmd, typeof(EventCounterResp));

            return resp.eventCounter;
        }

        public void SetGpioChipSettings()
        {
            GpioSettings cmd = new GpioSettings()
            {
                cmdCode = Commands.SetGpioChipSettings,
                gpioFunctions = new GpioFunction[9],
                gpioValue = 0,
                gpioDirection = 0,
                otherChipSettings = 0,
                chipParmAccessControl = 0,
            };

            ushort dir = 0;
            ushort output = 0;

            for (int i = 0; i < cmd.gpioFunctions.Length; i++)
            {
                cmd.gpioFunctions[i] = gpioSettings.gpioFunctions[i];
                if (Gpios[i].Direction)
                    dir |= (ushort)(1 << i);
                if (Gpios[i].Value)
                    output |= (ushort)(1 << i);
            }

            cmd.gpioDirection = dir;
            cmd.gpioValue = output;
            cmd.otherChipSettings = gpioSettings.otherChipSettings;

            cmd = (GpioSettings)sendCommand(cmd, typeof(GpioSettings));
            if(cmd.success != 0x00)
            {
                throw new Exception("Could not set GPIO settings");
            }
            //gpioSettings = cmd;

        }

        public void GetGpioChipSettings()
        {
            GetGpioSettingsReq req = new GetGpioSettingsReq()
            {
                cmdCode = Commands.GetGpioChipSettings,
            };

            GpioSettings resp = (GpioSettings)sendCommand(req, typeof(GpioSettings));
            if(resp.success != 0x00)
            {
                throw new Exception("Could not retreive GPIO settings");
            }
            gpioSettings = resp;

        }

        public void SetGpioValues()
        {
            GpioValues cmd = new GpioValues()
            {
                cmdCode = Commands.SetGpioCurrentPinValue,
            };
            cmd.pinValues = gpioSettings.gpioValue;

            cmd = (GpioValues)sendCommand(cmd, typeof(GpioValues));
            if(cmd.success != 0x00)
            {
                throw new Exception("Could not set GPIO values");
            }
        }

        public void GetGpioValues()
        {
            GetGpioSettingsReq cmd = new GetGpioSettingsReq() // this is not quite nice (repurposing other command)
            {
                cmdCode = Commands.GetGpioCurrentPinValue,
            };

            GpioValues resp = (GpioValues)sendCommand(cmd, typeof(GpioValues));

            if(resp.success != 0x00)
            {
                throw new Exception("Could not get GPIO values");
            }

            gpioSettings.gpioValue = resp.pinValues;
        }

        public void SetGpioDirections()
        {
            GpioValues cmd = new GpioValues()
            {
                cmdCode = Commands.SetGpioCurrentPinDirection,
            };

            cmd.pinValues = gpioSettings.gpioDirection;

            cmd = (GpioValues)sendCommand(cmd, typeof(GpioValues));
            if (cmd.success != 0x00)
            {
                throw new Exception("Could not set GPIO directions");
            }
        }

        public void GetGpioDirections()
        {
            GetGpioSettingsReq cmd = new GetGpioSettingsReq() // this is not quite nice (repurposing other command)
            {
                cmdCode = Commands.GetGpioCurrentPinDirection,
            };

            GpioValues resp = (GpioValues)sendCommand(cmd, typeof(GpioValues));

            if (resp.success != 0x00)
            {
                throw new Exception("Could not get GPIO directions");
            }

            gpioSettings.gpioDirection = resp.pinValues;
        }

        private void getSpiSettings()
        {
            GetSpiTransferSettingsReq req = new GetSpiTransferSettingsReq()
            {
                cmdCode = Commands.GetSpiTransferSettings,
            };

            SpiTransferSettings resp = (SpiTransferSettings)sendCommand(req, typeof(SpiTransferSettings));
            if(resp.success != 0x00)
            {
                throw new Exception("Could not get SPI settings");
            }
            spiSettings = resp;
        }

        private void setSpiSettings()
        {
            spiSettings.cmdCode = Commands.SetSpiTransferSettings;

            SpiTransferSettings resp = (SpiTransferSettings)sendCommand(spiSettings, typeof(SpiTransferSettings));
            if (resp.success != 0x00)
            {
                throw new Exception("Could not set SPI settings");
            }
            spiSettings = resp;
        }

        private void setSpiSettings(Mcp2210SpiChannel channel, ushort datalength)
        {
            spiSettings.cmdCode = Commands.SetSpiTransferSettings;
            spiSettings.bitrate = channel.Bitrate;
            spiSettings.delayCsToData = channel.DelayCsToData;
            spiSettings.delayDataToCs = channel.DelayDataToCs;
            spiSettings.delayDataToData = channel.DelayDataToData;
            spiSettings.spiMode = channel.SpiMode;

            spiSettings.idleCsValues = 0;
            spiSettings.activeCsValues = 0;

            spiSettings.bytesPerTransfer = (ushort)datalength;

            for (int i = 0; i < Gpios.Length; i++)
            {
                if (channel != null && Gpios[i] == channel.ChipSelectPin)
                {
                    if (Gpios[i].ChipSelectIdle == true)
                        spiSettings.idleCsValues |= (ushort)(1 << i);
                    if (Gpios[i].ChipSelectActive == true)
                        spiSettings.activeCsValues |= (ushort)(1 << i);
                }
                else
                {
                    if (Gpios[i].ChipSelectIdle == true)
                    {
                        spiSettings.idleCsValues |= (ushort)(1 << i);
                        spiSettings.activeCsValues |= (ushort)(1 << i);
                    }
                }
            }
            
            SpiTransferSettings resp = (SpiTransferSettings)sendCommand(spiSettings, typeof(SpiTransferSettings));
            if(resp.success != 0x00)
            {
                throw new Exception("Could not set SPI settings");

            }

            spiSettings = resp;
        }



        /// <summary>
        /// Transfers data over the given SPI channel. Rather use the method in Mcp2210SpiChannel
        /// </summary>
        /// <param name="channel">Mcp2210SpiChannel that's used for the transmission</param>
        /// <param name="data">Data bytes to be transmitted, maximum 65535 bytes</param>
        /// <returns>Data bytes received from the device</returns>
        public byte[] Transfer(Mcp2210SpiChannel channel, byte[] data)
        {
            if(data.Length > ushort.MaxValue)
            {
                throw new Exception("Transfer data length too long");
            }

            //chip select pin and state

            CancelSpiResp cancelresp = SpiCancel();
            if(cancelresp.success != 0x00)
            {
                throw new Exception("SPI channel could not be reset");
            }
            
            setSpiSettings(channel, (ushort)data.Length);

            int tx_offset = 0;
            int rx_offset = 0;
            int datalen = data.Length;

            byte[] rx_data = new byte[datalen];

            TransferSpiDataCmd cmd = new TransferSpiDataCmd()
            {
                cmdCode = Commands.TransferSpiData,
                data = new byte[60],
            };

            int cnt = 0;

            while (tx_offset != datalen || rx_offset != datalen)
            {
                int tx_blocklen = Math.Min(60, datalen - tx_offset);
                cmd.transmitLength = (byte)tx_blocklen;
                if (tx_blocklen > 0)
                    Array.Copy(data, tx_offset, cmd.data, 0, tx_blocklen);

                TransferSpiDataResp resp = (TransferSpiDataResp)sendCommand(cmd, typeof(TransferSpiDataResp));

                if(resp.respCode == 0xF7)
                {
                    // bus owned by external master
                    throw new Exception("Bus owned by external master");
                }
                else if(resp.respCode == 0x00)
                {
                    // transfer was successful
                    tx_offset += tx_blocklen;

                    if (resp.receiveLength > 0)
                    {
                        Array.Copy(resp.data, 0, rx_data, rx_offset, resp.receiveLength);
                    }
                    rx_offset += resp.receiveLength;
                    cnt = 0;
                }
                else if(resp.respCode == 0xF8)
                {
                    cnt++;
                }
                
                if(cnt > 300)
                {
                    throw new Exception("Too many retries for transmission");
                }
            }

            return rx_data;
        }

        /// <summary>
        /// Cancels an ongoing SPI transmission
        /// </summary>
        /// <returns>Status of the command</returns>
        public CancelSpiResp SpiCancel()
        {
            AnyCmd cmd = new AnyCmd()
            {
                cmdCode = Commands.TransferSpiCancel,
            };
            return (CancelSpiResp)sendCommand(cmd, typeof(CancelSpiResp));
        }

        /// <summary>
        /// Sets up GP6 for Interrupt counting
        /// </summary>
        /// <param name="irqmode">Mode how interrupts being counted</param>
        public void SetInterruptCounting(InterruptMode irqmode)
        {
            bool tmpimmediate = Immediate;
            Immediate = false;
            Gpios[6].Function = GpioFunction.DedicatedFunction;
            Gpios[6].Direction = true; // necessary?
            Gpio6IrqMode = irqmode;
            SetGpioChipSettings();
            Immediate = tmpimmediate;
        }


        #region Internal helper methods for subclasses
        internal void _setGpioFunctions()
        {
            if (Immediate == false) return;

            SetGpioChipSettings();
        }

        internal void _getGpioFunctions()
        {
            if (Immediate == false) return;

            GetGpioChipSettings();
        }

        internal void _setGpioValues()
        {
            if (Immediate == false) return;

            SetGpioValues();
        }

        internal void _getGpioValues()
        {
            if (Immediate == false) return;

            GetGpioValues();
        }

        internal void _setGpioDirections()
        {
            if (Immediate == false) return;
            SetGpioDirections();
        }

        internal void _getGpioDirections()
        {
            if (Immediate == false) return;

            GetGpioDirections();
        }
        #endregion
    }

    public class Mcp2210Gpio
    {
        private Mcp2210Device mcp = null;
        private byte index;

        /// <summary>
        /// Gets or sets the function of the GPIO.
        /// When Immediate of Mcp2210Device is false, [GS]etGpioChipSettings must be called to retreive/apply the value.
        /// </summary>
        public Mcp2210Device.GpioFunction Function
        {
            get
            {
                mcp._getGpioFunctions();
                return mcp.gpioSettings.gpioFunctions[index];
            }
            set
            {
                mcp.gpioSettings.gpioFunctions[index] = value;
                mcp._setGpioFunctions();
            }
        }

        /// <summary>
        /// Sets the direction of a GPIO. False is output, true is input
        /// When Immediate of Mcp2210Device is false, [GS]etGpioValues must be called to retreive/apply the value.
        /// </summary>
        public bool Direction
        {
            get
            {
                mcp._getGpioDirections();
                return (mcp.gpioSettings.gpioDirection & (1 << index)) != 0;
            }
            set
            {
                ushort tmp = mcp.gpioSettings.gpioDirection;
                if(value)
                {
                    tmp |= (ushort)(1 << index);
                }
                else
                {
                    tmp &= (ushort)~(1 << index);
                }
                mcp.gpioSettings.gpioDirection = tmp;
                mcp._setGpioDirections();
            }
        }

        /// <summary>
        /// Sets the direction of a GPIO.
        /// When Immediate of MCp2210Device is false, [GS]etGpioValues must be called to retreive/apply the set value.
        /// </summary>
        public bool Value
        {
            get
            {
                mcp._getGpioValues();
                return (mcp.gpioSettings.gpioValue & (1 << index)) != 0;
            }
            set
            {
                ushort tmp = mcp.gpioSettings.gpioValue;
                if (value)
                {
                    tmp |= (ushort)(1 << index);
                }
                else
                {
                    tmp &= (ushort)~(1 << index);
                }
                mcp.gpioSettings.gpioValue = tmp;
                mcp._setGpioValues();
            }
        }

        /// <summary>
        /// Sets or gets the GPIOs value for ChipSelect when it's in idle state
        /// </summary>
        public bool ChipSelectIdle = true;

        /// <summary>
        /// Sets or gets the GPIOs value for ChipSelect when it's in active state
        /// </summary>
        public bool ChipSelectActive = false;

        public Mcp2210Gpio(Mcp2210Device device, byte index)
        {
            mcp = device;
            this.index = index;
        }
    }

    public class Mcp2210SpiChannel
    {
        /// <summary>
        /// Sets or gets the SPI's bitrate in bits/second. Value is quite coarse.
        /// </summary>
        public UInt32 Bitrate = 100000;

        /// <summary>
        /// Sets or gets the SPI mode for the communication
        /// </summary>
        public Mcp2210Device.SpiMode SpiMode = Mcp2210Device.SpiMode.Mode0;

        /// <summary>
        /// Sets or gets the GPIO that's being used as Chip Select
        /// </summary>
        public Mcp2210Gpio ChipSelectPin;

        /// <summary>
        /// Wait time between CS gets active and first data byte is sent in 100 µs steps. Value is quite coarse.
        /// </summary>
        public ushort DelayCsToData = 0;
        /// <summary>
        /// Wait time between data bytes in 100 µs steps. Value is quite coarse.
        /// </summary>
        public ushort DelayDataToCs = 0;
        /// <summary>
        /// Wait time between last data byte is fully transmitted and CSs gets inactive in 100 µs steps. Value is quite coarse.
        /// </summary>
        public ushort DelayDataToData = 0;

        private Mcp2210Device mcp = null;

        public Mcp2210SpiChannel(Mcp2210Device device, Mcp2210Gpio ChipSelectPin)
        {
            mcp = device;
            this.ChipSelectPin = ChipSelectPin;
            this.ChipSelectPin.Function = Mcp2210Device.GpioFunction.ChipSelect;
            mcp.SetGpioChipSettings();
        }

        /// <summary>
        /// Transfers data bidirectionally over SPI
        /// </summary>
        /// <param name="data">Data bytes to be transmitted, maximum 65535 bytes</param>
        /// <returns>Data bytes received from the device</returns>
        public byte[] Transfer(byte[] data)
        {
            return mcp.Transfer(this, data);
        }
    }
}
