﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;

namespace Mcp2221
{
    public partial class Mcp2221Device
    {
        #region Enums

        private enum Commands : byte
        {
            StatusSetParameters = 0x10,
            ReadFlashData = 0xB0,
            WriteFlashData = 0xB1,
            SendFlashAccessPassword = 0xB2,

            I2cWriteData = 0x90,
            I2cWriteDataRepeatedStart = 0x92,
            I2cWriteDataNoStop = 0x94,
            I2cReadData = 0x91,
            I2cReadDataRepeatedStart = 0x93,
            I2cGetReadData = 0x40,

            SetGpioOutputValues = 0x50,
            GetGpioValues = 0x51,

            SetSramSettings = 0x60,
            GetSramSettings = 0x61,

            ResetChip = 0x70,
        }

        public enum I2cFraming : byte
        {
            Normal = 0x90,
            RepeatedStart = 0x92,
            NoStop = 0x94,
        }

        public enum ClockOutputDivider : byte
        {
            Cd_2 = 1,   //  24 MHz
            Cd_4 = 2,   //  12 MHz
            Cd_8 = 3,   //   6 MHz
            Cd_16 = 4,  //  3 MHz
            Cd_32 = 5,  //  1.5 MHz
            Cd_64 = 6,  //    750 kHz
            Cd_128 = 7, //    375 kHz
        }

        public enum ClockOutputDutycycle : byte
        {
            Duty_00 = 0,
            Duty_25 = 1,
            Duty_50 = 2,
            Duty_75 = 3,
        }

        public enum AnalogReference : byte
        {
            Vrm4096mV = 0b111,
            Vrm2048mV = 0b101,
            Vrm1024mV = 0b011,
            VrmOff = 0b001,
            Vdd = 0b000, // or any other with the last bit 0
        }

        /// <summary>
        /// This is mostly guesswork as Microchip didn't want to tell me the internal states.
        /// That's also why the code uses magic numbers instead of the enum.
        /// Some values can be found in MCP's linux driver sources (i2c-mcp2221.c)
        /// </summary>
        private enum I2cInternalState : byte
        {
            Idle = 0x00,
            CmdExecuted = 0x10, // unsure
            StartTimeout = 0x12,
            RestartTimeout = 0x17,
            WrAddrRlWsend = 0x21,
            WrAddrRLTimeout = 0x23,
            WriteAddrNak = 0x25,
            WriteDataTimeout = 0x44,
            WriteAddrDataAck = 0x45,
            DataReadOngoing = 0x50,
            ReadDataTimeout = 0x52,
            DataReadComplete = 0x55,
            DataReadPending = 0x54,
            Locked = 0x61, // i2c engine is in a lock state, no clue how to get out of it
            StopTimeout = 0x62,
            ReadError = 0x7F,
        }

        public enum InterruptConfig : byte
        {
            Off = 0,
            Rising = 1,
            Falling = 2,
            Both = 3,
        }
        #endregion

        #region Communication Structs
        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct StatusCmd
        {
            public byte cmdCode; // always 0x10
            private byte dnc0;
            public byte cancelI2cTransmission;
            public byte setI2cSpeed;
            public byte i2cClockDivider;
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct StatusResp
        {
            public byte cmdCode;
            public byte success;
            public byte cancelTransfer;
            public byte I2cSpeedCmd;
            public byte I2cClockDividerCmd;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 3)]
            private byte[] dnc0;
            public byte I2cIntState;
            public ushort I2cTransferLength;
            public ushort I2cTransferDone;
            public byte I2cIntBuffCnt;
            public byte I2cClockDivider;
            public byte I2cTimeoutValue;
            public ushort I2cAddress;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)] //18 ... 21
            private byte[] dnc1;
            public byte I2cSclLineVal;
            public byte I2cSdaLineVal;
            public byte IrqState;
            public byte I2cReadPending; //25
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 20)] //26 ... 45
            private byte[] dnc2;
            public ushort HwRev; //46 ... 47
            public ushort FwRev; //48 ... 49
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 3)]
            public ushort[] AdcVal; // 50 ... 51
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 6)]
            private byte[] dnc3;
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct I2cWriteDataCmd
        {
            public byte cmdCode; // 0x90
            public ushort transferLength;
            public byte i2cAddress;
            [MarshalAs(UnmanagedType.ByValArray, ArraySubType = UnmanagedType.U1, SizeConst = 60)]
            public byte[] data;
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct I2cWriteDataResp
        {
            public byte cmdCode;
            public byte success;
            public byte I2cIntState;
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct I2cReadDataCmd
        {
            public byte cmdCode;
            public ushort transferLength;
            public byte i2cAddress;
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct I2cReadDataResp
        {
            public byte cmdCode;
            public byte success;
            public byte I2cIntState;
            [MarshalAs(UnmanagedType.ByValArray, ArraySubType = UnmanagedType.U1, SizeConst = 61)]
            public byte[] data;
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct I2cGetDataCmd
        {
            public byte cmdCode; // always 0x40
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct I2cGetDataResp
        {
            public byte cmdCode;
            public byte success;
            public byte I2cIntState;
            public byte dataLength;
            [MarshalAs(UnmanagedType.ByValArray, ArraySubType = UnmanagedType.U1, SizeConst = 60)]
            public byte[] data;

        }

        //[StructLayout(LayoutKind.Sequential, Size = 4, Pack = 1)]
        public struct GpioSetOutValuesElement
        {
            public byte AlterGpioOutput;
            public byte GpioOutput;
            public byte AlterGpioDirection;
            public byte GpioDirection;
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct GpioSetOutValues
        {
            public byte cmdCode;
            private byte dnc0;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
            public GpioSetOutValuesElement[] Gpio;
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct GpioGetValuesCmd
        {
            public byte cmdCode;
        }

        [StructLayout(LayoutKind.Sequential, Size = 2, Pack = 1)]
        public struct GpioGetValuesRespElement
        {
            public byte PinValue;
            public byte PinDirection;
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct GpioGetValuesResp
        {
            public byte cmdCode;
            public byte success;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
            public GpioGetValuesRespElement[] Gpio;
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct SramSetSettingsCmd
        {
            public byte cmdCode; // 0x60
            private byte dnc0;
            public byte ClkOutput;
            public byte DacVref;
            public byte DacValue;
            public byte AdcVref;
            public byte InterruptSettings;
            public byte AlterGpios;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
            public byte[] GpioSettings;
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct SramSetSettingsResp
        {
            public byte cmdCode;
            public byte success;
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct SramGetSettingsCmd
        {
            public byte cmdCode; // 0x61
        }

        [StructLayout(LayoutKind.Sequential, Size = 64, Pack = 1)]
        public struct SramGetSettingsResp
        {
            public byte cmdCode;
            public byte success;
            public byte ChipSettingsLength;
            public byte GpioSettingsLength;
            public byte GenSettings;
            public byte ClockOutputDivider;
            public byte DacConfig;
            public byte Irq_Adc;
            public ushort UsbVid;
            public ushort UsbPid;
            public byte UsbPowerAttributes;
            public byte UsbCurrent;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
            public byte[] SecurityPassword;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
            public byte[] GpioSettings;
        }
        #endregion
    }
}
