﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using SimplerHid;

namespace Mcp2221
{

    public partial class Mcp2221Device
    {

        #region Constants
        private static readonly double I2cClkdivOffset = -5.5f;
        private static readonly double I2cClkdivBasefreq = 12e6;
        private static readonly byte I2cClkdivMin = 20;
        private static readonly byte I2cClkdivMax = 255;
        #endregion

        #region Helper functions
        private float Adc_RefVoltage(AnalogReference reference, int adcValue)
        {
            switch (reference)
            {
                case AnalogReference.Vrm4096mV: return 4.096f;
                case AnalogReference.Vrm2048mV: return 2.048f;
                case AnalogReference.Vrm1024mV: return 1.024f;
                case AnalogReference.Vdd: return 5.0f; // can be more, can be less, let's assume it's 5 V
                default:
                    throw new Exception("Voltage can't be caluclated");
            }
        }

        private static byte i2c_calcClkdiv(double frequency)
        {
            int clkdiv = (int)Math.Round(I2cClkdivBasefreq / frequency + I2cClkdivOffset);
            return (byte)Math.Max(I2cClkdivMin, Math.Min(I2cClkdivMax, clkdiv));
        }

        private static double i2c_calcFreq(byte clkdiv)
        {
            return (double)(I2cClkdivBasefreq / ((double)clkdiv - I2cClkdivOffset));
        }

        private static string arr2hexstr(byte[] arr, string prefix = "", string separator = " ")
        {
            StringBuilder sb = new StringBuilder();
            foreach (byte item in arr)
            {
                sb.Append(prefix);
                sb.AppendFormat("{0:X2}", item);
                sb.Append(separator);
            }
            if(sb.Length > 0)
            {
                sb.Length = sb.Length - separator.Length;
            }
            return sb.ToString();
        }
        #endregion

        #region Variables
        private HidDevice device = null;
        private StatusResp status;
        public Mcp2221Gpio[] Gpios = null;

        public Mcp2221Gpio0 Gpio0;
        public Mcp2221Gpio1 Gpio1;
        public Mcp2221Gpio2 Gpio2;
        public Mcp2221Gpio3 Gpio3;

        InterruptConfig interruptConfig;
        private ClockOutputDivider clockOutputDivider;
        private ClockOutputDutycycle clockOutputDutycycle;
        private byte dacValue;
        private AnalogReference dacReference;
        private AnalogReference adcReference;
        //private byte i2cClockDiv;

        //private bool paramsdirty = true;
        //private bool iodirty = true;
        //private bool iocfgdirty = true;
        #endregion

        /// <summary>
        /// Defines if information is read/write immediately from/to the device.
        /// This is to speed up communication when multiple operations (like setting GPIOs or reading ADCs) are done consecutively.
        /// If false, each read/write access will run through a full communication cycle of around 2 ms
        /// </summary>
        public bool Immediate = true;

        public Mcp2221Device(string devicePath)
        {
            device = new HidDevice(devicePath);
            
            // yep, the following pretty much sucks but I couldn't find anything better than that (yet)
            Gpio0 = new Mcp2221Gpio0(this);
            Gpio1 = new Mcp2221Gpio1(this);
            Gpio2 = new Mcp2221Gpio2(this);
            Gpio3 = new Mcp2221Gpio3(this);
            Gpios = new Mcp2221Gpio[]
            {
                Gpio0,
                Gpio1,
                Gpio2,
                Gpio3,
            };

            // read all info
            SramReadSettings();
            I2cSetParameter();
        }

        public void Dispose()
        {
            device.Dispose();
        }

        public double I2cSpeed
        {
            get
            {
                if (Immediate == true)
                    I2cGetStatus();
                return i2c_calcFreq(status.I2cClockDivider);
            }
            set
            {
                status.I2cClockDivider = i2c_calcClkdiv(value);
                if (Immediate == true)
                    I2cSetParameter(false, status.I2cClockDivider);
            }
        }

        public AnalogReference AdcReference
        {
            get
            {
                if (Immediate)
                    SramReadSettings();
                return adcReference;
            }
            set
            {
                adcReference = value;
                if (Immediate)
                    AdcSet();
            }
        }

        public AnalogReference DacReference
        {
            get
            {
                if (Immediate)
                    SramReadSettings();
                return dacReference;
            }
            set
            {
                dacReference = value;
                if (Immediate)
                    DacSet();
            }
        }

        private object sendCommand(object cmd, Type respType)
        {
            byte[] tmp = HidDevice.GetBytesFromObject(cmd);
            //Console.WriteLine("> " + arr2hexstr(tmp));
            device.WriteOutputReport(0, tmp);
            byte[] resp = device.ReadInputReport();
            //Console.WriteLine("< " + arr2hexstr(resp));
            //Console.WriteLine();
            return HidDevice.GetObjectFromBytes(resp, respType);
        }

        private StatusResp I2cSetParameter(bool i2cCancel = false, byte? i2cSpeedDiv = null)
        {
            StatusCmd cmd = new StatusCmd()
            {
                cmdCode = (byte)Commands.StatusSetParameters,
                cancelI2cTransmission = (byte)((i2cCancel == true) ? 0x10 : 0x00),
            };

            if(i2cSpeedDiv != null)
            {
                cmd.setI2cSpeed = 0x20;
                cmd.i2cClockDivider = (byte)i2cSpeedDiv;
            }

            //device.WriteOutputReport(0, HidDevice.GetBytesFromObject(cmd));
            //this.status = (StatusResp)HidDevice.GetObjectFromBytes(device.ReadInputReport(), typeof(StatusResp));

            this.status = (StatusResp)sendCommand(cmd, typeof(StatusResp));

            return this.status;
        }

        public void I2cCancel()
        {
            I2cSetParameter(true);
        }

        private void I2cGetStatus()
        {
            I2cSetParameter();
        }

        public bool I2cWrite(byte address, byte[] data, I2cFraming framing = I2cFraming.Normal)
        {

            if (data.Length > 0xFFFF)
            {
                throw new Exception("data Length > 0xFFFF is not supported");
            }

            I2cWriteDataCmd cmd = new I2cWriteDataCmd()
            {
                cmdCode = (byte)framing,
                i2cAddress = (byte)((address << 1) & 0xFF),
                transferLength = (byte)data.Length,
                data = new byte[60],
            };

            int slicelen = 60;
            int slices = (data.Length + slicelen - 1) / slicelen;
            slices = Math.Max(slices, 1);
            int status = -1;

            int retrans_max = 20;
            int retrans_tot = 0;

            for(int s = 0; s < slices; s++)
            {
                Array.Copy(data, s * 60, cmd.data, 0, Math.Min(60, data.Length - slicelen * s));

                for (int r = 0; r <= retrans_max; r++)
                {
                    I2cWriteDataResp resp = (I2cWriteDataResp)sendCommand(cmd, typeof(I2cWriteDataResp));
                    status = resp.success;

                    this.status.I2cIntState = resp.I2cIntState;

                    if(resp.success == 0)
                        break;

                    retrans_tot++;

                    if (r == retrans_max)
                    {
                        I2cCancel();
                        throw new Exception("Retransmission limit reached");
                    }
                    System.Threading.Thread.Sleep(2);
                }

                if(status != 0)
                {
                    I2cCancel();
                    return false;
                }
                
            }
            return true;
        }

        public byte[] I2cRead(byte address, int length, bool repeated_start = false)
        {
            if (length <= 0)
                throw new Exception("I2C read data length <= 0 is not supported");
            else if(length > 0xFFFF)
                throw new Exception("I2C read data length > 0xFFFF is not supported");

            I2cReadDataCmd cmd = new I2cReadDataCmd()
            {
                cmdCode = (byte)(repeated_start == true ?  Commands.I2cReadDataRepeatedStart : Commands.I2cReadData),
                i2cAddress = (byte)((address << 1) | 1),
                transferLength = (ushort)length,
            };

            I2cReadDataResp resp = (I2cReadDataResp)sendCommand(cmd, typeof(I2cReadDataResp));

            if (resp.success != 0)
                throw new Exception("unable to read data");

            I2cGetStatus();
            if (this.status.I2cIntState == 0x25)
                throw new Exception("Got NAK to read request");

            return i2cGetData();
        }

        private byte[] i2cGetData()
        {
            //List<byte> data = new List<byte>();
            MemoryStream stream = new MemoryStream();

            I2cGetDataCmd cmd = new I2cGetDataCmd()
            {
                cmdCode = (byte)Commands.I2cGetReadData,
            };

            int error_waterlevel = 0;
            I2cGetDataResp resp;

            while (true)
            {
                resp = (I2cGetDataResp)sendCommand(cmd, typeof(I2cGetDataResp));

                this.status.I2cIntState = resp.I2cIntState;

                if (resp.success == 0)
                {
                    if (error_waterlevel > 0)
                        error_waterlevel--;
                }
                else if (this.status.I2cIntState != 0x50)
                    error_waterlevel++;

                if (error_waterlevel > 200)
                    throw new Exception("too many unsuccessful read attempts");

                if (this.status.I2cIntState == 0x25)
                    throw new Exception("write addr NAK");

                if (resp.dataLength == 127)
                    continue;
                else if (resp.dataLength > 0 && resp.dataLength <= 60)
                {
                    //byte[] tmp = new byte[resp.dataLength]
                    stream.Write(resp.data, 0, resp.dataLength);
                }

                if(this.status.I2cIntState == 0x55)
                {
                    //stream.Flush();
                    return stream.ToArray();
                }
            }
        }

        public byte[] I2cReadRegister8(byte address, byte register, int readLength)
        {
            if (!I2cWrite(address, new byte[] { register }, I2cFraming.NoStop))
                return null;

            byte[] data = I2cRead(address, readLength, true);
            I2cGetStatus();
            return data;
        }

        public bool I2cWriteRegister8(byte address, byte register, byte[] data)
        {
            byte[] tmp = new byte[data.Length + 1];
            tmp[0] = register;
            Array.Copy(data, 0, tmp, 1, data.Length);
            return I2cWrite(address, tmp, I2cFraming.Normal);
        }

        public List<byte> I2cDetect(byte addrStart = 0x00, byte addrStop = 0x7F)
        {
            byte addr_f = new byte[] { addrStart, addrStop, 0x7F }.Min();
            byte addr_t = (byte)(new byte[] { addrStart, addrStop, 0 }.Max() + 1);

            List<byte> found = new List<byte>();

            for(byte a = addr_f; a < addr_t; a++)
            {
                //Debug.WriteLine("Probing address 0x{0:X2}", a);
                I2cWrite(a, new byte[0]);
                I2cGetStatus();
                if (status.I2cIntState == 0)
                    found.Add(a);
                else
                    I2cCancel();
                
            }

            return found;
        }

        public void GpioSet()
        {
            GpioSetOutValues cmd = new GpioSetOutValues()
            {
                cmdCode = (byte)Commands.SetGpioOutputValues,
                Gpio = new GpioSetOutValuesElement[4],
            };
            for (int i = 0; i < cmd.Gpio.Length; i++)
            {
                cmd.Gpio[i] = Gpios[i].dev_SetGpioOutputValues();
            }

            sendCommand(cmd, typeof(GpioSetOutValues)); // TODO: process return values

        }

        public bool GpioGet()
        {
            GpioGetValuesCmd cmd = new GpioGetValuesCmd()
            {
                cmdCode = (byte)Commands.GetGpioValues,
            };

            GpioGetValuesResp resp = (GpioGetValuesResp)sendCommand(cmd, typeof(GpioGetValuesResp));

            if (resp.success != 0)
                return false;

            for (int i = 0; i < Gpios.Length; i++)
            {
                Gpios[i]._dev_GetGpioOutputValues(resp.Gpio[i]);
            }

            return true;
        }

        public void Reset()
        {
            byte[] cmd = new byte[64];
            Array.Copy(new[] { 0x70, 0xAB, 0xCD, 0xEF }, 0, cmd, 0, 4);
            device.WriteOutputReport(0, cmd);
            //TODO: reconnect
        }

        public bool AdcSet()
        {
            return SramWriteSettings(adcReference: adcReference);
        }

        public bool DacSet()
        {
            return SramWriteSettings(dacReference: dacReference, dacValue: dacValue);
        }

        public bool WriteConfig()
        {
            return SramWriteSettings(
                clockOutputDutycycle, clockOutputDivider, 
                dacReference, dacValue, 
                adcReference, 
                interruptConfig, false, 
                true);
        }

        private bool SramWriteSettings(ClockOutputDutycycle? dutycycle = null, ClockOutputDivider? divider = null,
            AnalogReference? dacReference = null, byte? dacValue = null,
            AnalogReference? adcReference = null,
            InterruptConfig? interruptCfg = null, bool irqClear = false,
            bool setGpios = false
            )
        {
            
            SramSetSettingsCmd cmd = new SramSetSettingsCmd()
            {
                cmdCode = (byte)Commands.SetSramSettings,
            };

            if(dutycycle != null && divider != null)
            {
                cmd.ClkOutput = (byte)((1 << 7) | (((byte)dutycycle & 0x03) << 3) | (byte)divider & 0x07);
            }
            else if(dutycycle == null && divider == null)
            {
                //do nothing
            }
            else
            {
                throw new Exception("CLKOUT duty and divider must be set at the same time");
            }

            if(dacReference != null)
            {
                cmd.DacVref = (byte)((1 << 7) | (byte)dacReference);
                this.dacReference = (AnalogReference)dacReference;
            }

            if(dacValue != null)
            {
                cmd.DacValue = (byte)((1 << 7) | (dacValue & 0x1F));
            }

            if (adcReference != null)
            {
                cmd.AdcVref = (byte)((1 << 7) | (byte)adcReference);
                this.adcReference = (AnalogReference)adcReference;
            }

            byte tmp = 0;
            if(interruptCfg != null)
            {
                tmp |= (1 << 7) | (1 << 4) | (1 << 2);
                switch (interruptCfg)
                {
                    case InterruptConfig.Rising:
                        tmp |= (1 << 3);
                        break;
                    case InterruptConfig.Falling:
                        tmp |= (1 << 1);
                        break;
                    case InterruptConfig.Both:
                        tmp |= (1 << 3) | (1 << 1);
                        break;
                }
            }

            if (irqClear == true)
                tmp |= (1 << 7) | (1 << 0);

            cmd.InterruptSettings = tmp;

            if(setGpios)
            {
                cmd.AlterGpios = 0x80; // the datasheet is wrong here
                cmd.GpioSettings = new byte[Gpios.Length];
                for (int i = 0; i < Gpios.Length; i++)
                {
                    cmd.GpioSettings[i] = Gpios[i]._dev_SetSramSettings();
                }
            }

            SramSetSettingsResp resp = (SramSetSettingsResp)sendCommand(cmd, typeof(SramSetSettingsResp));

            return resp.success == 0x00;
        }

        public void SramReadSettings()
        {
            SramGetSettingsCmd cmd = new SramGetSettingsCmd()
            {
                cmdCode = (byte)Commands.GetSramSettings,
            };

            SramGetSettingsResp resp = (SramGetSettingsResp)sendCommand(cmd, typeof(SramGetSettingsResp));

            //ignore first 4 bytes
            clockOutputDutycycle = (ClockOutputDutycycle)((resp.ClockOutputDivider >> 2) & 0x03);
            clockOutputDivider = (ClockOutputDivider)(resp.ClockOutputDivider & 0x07);

            if((resp.DacConfig & (1<<5)) == 0)
            {
                dacReference = AnalogReference.Vdd;
            }
            else
            {
                dacReference = (AnalogReference)(resp.DacConfig >> 5);
            }

            dacValue = (byte)(resp.DacConfig & 0x1F);

            interruptConfig = (InterruptConfig)((resp.Irq_Adc >> 5) & 0x03);

            if ((resp.Irq_Adc & (1 << 2)) == 0)
            {
                adcReference = AnalogReference.Vdd;
            }
            else
            {
                adcReference = (AnalogReference)((resp.Irq_Adc >> 2) & 0x07);
            }

            for (int i = 0; i < this.Gpios.Length; i++)
            {
                Gpios[i]._dev_GetSramSettings(resp.GpioSettings[i]);
            }
        }

        internal byte _dacGet(int index)
        {
            if (Immediate)
                SramReadSettings();
            return dacValue;
        }

        internal void _dacSet(int index, byte value)
        {
            dacValue = value;
            if (Immediate)
                SramWriteSettings(dacValue: value);
        }

        internal ushort _adcGet(int index)
        {
            if(Immediate)
                I2cSetParameter();
            return status.AdcVal[index-1];
        }

        internal void _gpioGet()
        {
            if (Immediate)
            {
                GpioGet();
                SramReadSettings();
            }
        }

        internal void _gpioSet()
        {
            if(Immediate)
            {
                //Sram_WriteSettings(setGpios: true);
                WriteConfig();
                GpioSet();
            }
            
        }

        internal void _gpioGetfunc()
        {
            if(Immediate)
            {
                SramReadSettings();
            }
        }

        internal void _gpioSetfunc()
        {
            if (Immediate)
            {
                SramWriteSettings(setGpios: true);
                //WriteConfig();
            }
        }


    }
}
