﻿// This code is copied from simplehidlibrary with some modifications
// for more simplicity by chris@hobbyelektronik.org

// This source is subject to Microsoft Public License (Ms-PL).
// Please see http://simplehidlibrary.codeplex.com/ for details.
// All other rights reserved.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Microsoft.Win32.SafeHandles;
using System.ComponentModel;
using System.Runtime.InteropServices;

namespace SimplerHid
{

    public class HIDInfoSet
    {
        /// <summary>
        /// Identifies the manufacturer's product
        /// </summary>
        public string ProductString { get; private set; }

        /// <summary>
        /// Serial number
        /// </summary>
        public string SerialNumberString { get; private set; }

        /// <summary>
        /// Full device file path
        /// </summary>
        public string DevicePath { get; private set; }

        /// <summary>
        /// Vendor ID
        /// </summary>
        public short Vid { get; private set; }

        /// <summary>
        /// Product ID
        /// </summary>
        public short Pid { get; private set; }

        /// <summary>
        /// Product Version
        /// </summary>
        public short Version { get; private set; }

        /// <summary>
        /// ctor
        /// </summary>
        internal HIDInfoSet(string productString, string serialNumberString, string devicePath, short vid, short pid, short version)
        {
            ProductString = productString;
            SerialNumberString = serialNumberString;
            DevicePath = devicePath;
            Vid = vid;
            Pid = pid;
            Version = version;
        }
    }

    public class HidDevice
    {
        private IntPtr handle;
        private FileStream fileStream;

        private IntPtr preparedData;
        private NativeMethods.HidCaps hidCaps;

        static public IEnumerable<HIDInfoSet> GetInfoSets()
        {
            var devicePathList = new List<HIDInfoSet>();
            Guid gHid;
            NativeMethods.HidD_GetHidGuid(out gHid);
            var hInfoSet = NativeMethods.SetupDiGetClassDevs(ref gHid, null, IntPtr.Zero, NativeMethods.DIGCF_DEVICEINTERFACE | NativeMethods.DIGCF_PRESENT);
            if (hInfoSet == NativeMethods.InvalidHandleValue)
            {
                throw new Win32Exception();
            }
            var oInterface = new NativeMethods.DeviceInterfaceData();
            oInterface.Size = Marshal.SizeOf(oInterface);
            int nIndex = 0;
            while (NativeMethods.SetupDiEnumDeviceInterfaces(hInfoSet, 0, ref gHid, (uint)nIndex, ref oInterface))
            {
                var strDevicePath = GetDevicePath(hInfoSet, ref oInterface);
                var handle = Open(strDevicePath);
                if (handle != NativeMethods.InvalidHandleValue)
                {
                    var productString = GetName(handle);
                    var serialNumberString = GetSerialNumber(handle);
                    NativeMethods.HIDD_ATTRIBUTES attributes;
                    GetAttr(handle, out attributes);
                    devicePathList.Add(new HIDInfoSet(productString, serialNumberString, strDevicePath, attributes.VendorID, attributes.ProductID, attributes.VersionNumber));
                }
                nIndex++;
            }
            if (NativeMethods.SetupDiDestroyDeviceInfoList(hInfoSet) == false)
            {
                throw new Win32Exception();
            }

            return devicePathList;
        }

        #region Private helper methods

        private static IntPtr Open(string devicePath)
        {
            return NativeMethods.CreateFile(devicePath, NativeMethods.GENERIC_READ | NativeMethods.GENERIC_WRITE, NativeMethods.FILE_SHARE_READ | NativeMethods.FILE_SHARE_WRITE, IntPtr.Zero, NativeMethods.OPEN_EXISTING, NativeMethods.FILE_FLAG_OVERLAPPED, IntPtr.Zero);
        }

        private static void GetAttr(IntPtr handle, out NativeMethods.HIDD_ATTRIBUTES attributes)
        {
            attributes = new NativeMethods.HIDD_ATTRIBUTES();
            attributes.Size = Marshal.SizeOf(attributes);
            if (NativeMethods.HidD_GetAttributes(handle, ref attributes) == false)
            {
                throw new Win32Exception();
            }
        }

        private static string GetName(IntPtr handle)
        {
            var manufacturerString = new StringBuilder(128);
            if (NativeMethods.HidD_GetProductString(handle, manufacturerString, manufacturerString.Capacity) == false)
            {
                throw new Win32Exception();
            }
            return manufacturerString.ToString();
        }

        private static string GetSerialNumber(IntPtr handle)
        {
            var serialNumberString = new StringBuilder(128);
            if (NativeMethods.HidD_GetSerialNumberString(handle, serialNumberString, serialNumberString.Capacity) == false)
            {
                // Some devices has no serial number
                // Just return an empty string
                return string.Empty;
            }
            return serialNumberString.ToString();
        }

        /// <summary>
        /// Get USB HID device path
        /// </summary>
        private static string GetDevicePath(IntPtr hInfoSet, ref NativeMethods.DeviceInterfaceData oInterface)
        {
            uint nRequiredSize = 0;
            if (NativeMethods.SetupDiGetDeviceInterfaceDetail(hInfoSet, ref oInterface, IntPtr.Zero, 0, ref nRequiredSize, IntPtr.Zero) == false)
            {
                // TODO: Find a solution
                // Tip: http://stackoverflow.com/questions/1054748/setupdigetdeviceinterfacedetail-unexplainable-error
                //throw new HIDDeviceException();
            }
            var oDetail = new NativeMethods.DeviceInterfaceDetailData();
            oDetail.Size = Marshal.SizeOf(typeof(IntPtr)) == 8 ? 8 : 5; // x86/x64 magic...
            if (NativeMethods.SetupDiGetDeviceInterfaceDetail(hInfoSet, ref oInterface, ref oDetail, nRequiredSize, ref nRequiredSize, IntPtr.Zero) == false)
            {
                throw new Win32Exception();
            }
            return oDetail.DevicePath;
        }

        #endregion

        public HidDevice(string devicePath)
        {

            handle = NativeMethods.CreateFile(devicePath, NativeMethods.GENERIC_READ | NativeMethods.GENERIC_WRITE, NativeMethods.FILE_SHARE_READ | NativeMethods.FILE_SHARE_WRITE, IntPtr.Zero, NativeMethods.OPEN_EXISTING, NativeMethods.FILE_FLAG_OVERLAPPED, IntPtr.Zero);
            
            if (NativeMethods.HidD_GetPreparsedData(handle, out preparedData))
            {

                if (NativeMethods.HidP_GetCaps(preparedData, out hidCaps) != NativeMethods.HIDP_STATUS_SUCCESS)
                {
                    throw new Win32Exception();
                }

                if (handle != NativeMethods.InvalidHandleValue || handle == null)
                {
                    fileStream = new FileStream(new SafeFileHandle(handle, false), FileAccess.Read | FileAccess.Write, hidCaps.InputReportByteLength, true);
                    return;
                }

            }

            throw new Exception("could not connect to device");
        }

        public void Dispose()
        {
            NativeMethods.HidD_FreePreparsedData(ref preparedData);

            fileStream.Close();
            fileStream = null;
            NativeMethods.CloseHandle(handle);
        }

        public void WriteOutputReport(byte reportId, byte[] data)
        {
            // should the data be prepared properly? I don't know...
            byte[] outdata = new byte[hidCaps.OutputReportByteLength];
            outdata[0] = reportId;
            Array.Copy(data, 0, outdata, 1, Math.Min(outdata.Length - 1, data.Length));
            fileStream.Write(outdata, 0, outdata.Length);
            fileStream.Flush();
        }

        public byte[] ReadInputReport(out byte reportId)
        {
            byte[] data = new byte[hidCaps.InputReportByteLength];
            /*if (NativeMethods.HidP_InitializeReportForID(NativeMethods.HIDP_REPORT_TYPE.HidP_Input, reportId, preparedData, data, data.Length) != NativeMethods.HIDP_STATUS_SUCCESS)
            {
                throw new Win32Exception();
            }*/
            fileStream.Read(data, 0, data.Length);

            reportId = data[0];

            byte[] data2 = new byte[hidCaps.InputReportByteLength - 1];
            Array.Copy(data, 1, data2, 0, data2.Length);

            return data2;
        }

        public byte[] ReadInputReport()
        {
            return ReadInputReport(out _);
        }

        #region Serializer functions
        public static object GetObjectFromBytes(byte[] buffer, Type objType)
        {
            object obj = null;
            if ((buffer != null) && (buffer.Length > 0))
            {
                IntPtr ptrObj = IntPtr.Zero;
                try
                {
                    int objSize = Marshal.SizeOf(objType);
                    if (objSize > 0)
                    {
                        if (buffer.Length < objSize)
                            throw new Exception(String.Format("Buffer smaller than needed for creation of object of type {0}", objType));
                        ptrObj = Marshal.AllocHGlobal(objSize);
                        if (ptrObj != IntPtr.Zero)
                        {
                            Marshal.Copy(buffer, 0, ptrObj, objSize);
                            obj = Marshal.PtrToStructure(ptrObj, objType);
                        }
                        else
                            throw new Exception(String.Format("Couldn't allocate memory to create object of type {0}", objType));
                    }
                }
                finally
                {
                    if (ptrObj != IntPtr.Zero)
                        Marshal.FreeHGlobal(ptrObj);
                }
            }
            return obj;
        }

        public static byte[] GetBytesFromObject(object obj)
        {
            // try/catch? for losers.
            int objSize = Marshal.SizeOf(obj.GetType());
            byte[] buffer = new byte[objSize];
            IntPtr ptrObj = IntPtr.Zero;
            ptrObj = Marshal.AllocHGlobal(objSize);
            Marshal.StructureToPtr(obj, ptrObj, false);
            Marshal.Copy(ptrObj, buffer, 0, objSize);
            Marshal.FreeHGlobal(ptrObj);

            return buffer;
        }
        #endregion

    }
}
