﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;


namespace heohelper
{
    [System.AttributeUsage(AttributeTargets.Property | AttributeTargets.Field, AllowMultiple = false)]
    public class BitfieldAttribute : Attribute
    {
        uint offset;
        uint length;

        public BitfieldAttribute(uint offset, uint length)
        {
            this.offset = offset;
            this.length = length;
        }
        public uint Length { get { return length; } }
        public uint Offset { get { return offset; } }
    }

    public static class Bitfield
    {

        public static byte[] GetBytes<T>(T t) where T : class
        {
            PropertyInfo[] fi = t.GetType().GetProperties();
            List<byte> bytes = new List<byte>();

            byte currByte = 0;
            int byteOffset = 0;

            foreach (PropertyInfo f in fi)
            {
                object[] attrs = f.GetCustomAttributes(typeof(BitfieldAttribute), false);

                if (attrs.Length != 1)
                    continue;

                //int fieldOffset = (int)((BitfieldAttribute)attrs[0]).Offset; //FIXME: OFFSET IS NOT SUPPORTED!
                int fieldLength = (int)((BitfieldAttribute)attrs[0]).Length;

                long mask = ((1 << fieldLength) - 1);
                ulong val = 0;


                object member = f.GetValue(t, null);
                if (member is byte || member is UInt16 || member is UInt32 || member is UInt64)
                    val = Convert.ToUInt64(member);
                else if (member is sbyte || member is Int16 || member is Int32 || member is Int64)
                    val = (ulong)unchecked(Convert.ToInt64(member));
                else if (member is bool)
                    val = (ulong)((bool)member == true ? 1 : 0);
                else if (f.GetValue(t, null).GetType().IsEnum)
                    val = unchecked((ulong)Convert.ToInt64(member));
                else
                    throw new Exception(string.Format("type {0} is not supported", f.GetType().ToString()));

                while (mask > 0)
                {
                    int bitsToInsert = Math.Min(fieldLength, (8 - byteOffset));
                    byte bmask = (byte)((1 << (bitsToInsert)) - 1);

                    currByte |= (Byte)((val & bmask) << byteOffset);
                    val >>= bitsToInsert;
                    mask >>= bitsToInsert;

                    byteOffset += bitsToInsert;

                    if (byteOffset == 8)
                    {
                        bytes.Add(currByte);
                        byteOffset = 0;
                        currByte = 0;
                    }
                }
            }

            if (byteOffset > 0)
            {
                bytes.Add(currByte);
            }

            return bytes.ToArray();
        }

        private static TypeCode[] signedTypes = new TypeCode[] { TypeCode.Int16, TypeCode.Int32, TypeCode.Int64, TypeCode.SByte };


        public static void SetBytes<T>(byte[] bytes, T t) where T : class
        {
            
            int offset = 0;
            PropertyInfo[] fi = t.GetType().GetProperties();
            foreach (var f in fi)
            {
                object[] attrs = f.GetCustomAttributes(typeof(BitfieldAttribute), false);

                if (attrs.Length != 1)
                    continue;

                int fieldLength = (int)((BitfieldAttribute)attrs[0]).Length;

                ulong part = BinaryPart(bytes, offset, fieldLength);
                offset += fieldLength;

                Type ft = f.PropertyType;

                if (ft.IsValueType && !ft.IsEnum)
                {
                    if (Array.IndexOf(signedTypes, Type.GetTypeCode(f.PropertyType)) != -1)
                    {
                        //item is signed, so we have to fill up the signed bit
                        //first we need to check if it's negative by the bit at the highest possible position
                        if (((part >> (fieldLength - 1)) & 1) == 1)
                        {
                            //it's actually negative, mask it out!
                            //this looks terribly stupid, but it works
                            //shifting left eliminates the 0s on the LSBs
                            part |= ulong.MaxValue << fieldLength;
                        }
                    }


                    switch (Type.GetTypeCode(f.PropertyType))
                    {
                        case TypeCode.Boolean: f.SetValue(t, (bool)(part != 0), null); break;
                        case TypeCode.Char: f.SetValue(t, unchecked((char)part), null); break;
                        case TypeCode.Byte: f.SetValue(t, (byte)part, null); break;
                        case TypeCode.UInt16: f.SetValue(t, unchecked((UInt16)part), null); break;
                        case TypeCode.UInt32: f.SetValue(t, unchecked((UInt32)part), null); break;
                        case TypeCode.UInt64: f.SetValue(t, unchecked((UInt64)part), null); break;
                        case TypeCode.SByte: f.SetValue(t, unchecked((sbyte)part), null); break;
                        case TypeCode.Int16: f.SetValue(t, unchecked((Int16)part), null); break;
                        case TypeCode.Int32: f.SetValue(t, unchecked((Int32)part), null); break;
                        case TypeCode.Int64: f.SetValue(t, unchecked((Int64)part), null); break;
                        //case TypeCode.Decimal:
                        //case TypeCode.Double:
                        //case TypeCode.Single:
                        //case TypeCode.String:
                        default:
                            throw new Exception("phuck! no so value type available.");
                    }
                }
                else if (ft.IsEnum)
                {
                    //this is very very very hacky
                    Int32 meh = unchecked((Int32)part);
                    if (Enum.IsDefined(ft, meh))
                    {
                        f.SetValue(t, meh, null);
                    }
                    else
                        throw new Exception("phuck! could not set enum :(");
                }
            }
        }


        /*public static byte[] GetBytes<T>(T t) where T : struct
        {
            PropertyInfo[] fi = t.GetType().GetProperties();
            List<byte> bytes = new List<byte>();

            byte currByte = 0;
            int byteOffset = 0;

            foreach (PropertyInfo f in fi)
            {
                object[] attrs = f.GetCustomAttributes(typeof(BitfieldAttribute), false);

                if (attrs.Length != 1)
                {
                    throw new Exception("no BitfieldLength-Attribute found");
                }

                int fieldLength = (int)((BitfieldAttribute)attrs[0]).Length;

                long mask = ((1 << fieldLength) - 1);
                ulong val = 0;

                
                object member = f.GetValue(t, null);
                if(member is byte || member is UInt16 || member is UInt32 || member is UInt64)
                    val = Convert.ToUInt64(member);
                else if (member is sbyte || member is Int16 || member is Int32 || member is Int64)
                    val = (ulong)unchecked(Convert.ToInt64(member));
                else if (member is bool)
                    val = (ulong)((bool)member == true ? 1 : 0);
                else if (f.GetValue(t, null).GetType().IsEnum)
                    val = unchecked((ulong)Convert.ToInt64(member));
                else
                    throw new Exception(string.Format("type {0} is not supported", f.GetType().ToString()));

                while (mask > 0)
                {
                    int bitsToInsert = Math.Min(fieldLength, (8 - byteOffset));
                    byte bmask = (byte)((1 << (bitsToInsert)) -1);

                    currByte |= (Byte)((val & bmask) << byteOffset);
                    val >>= bitsToInsert;
                    mask >>= bitsToInsert;

                    byteOffset += bitsToInsert;

                    if (byteOffset == 8)
                    {
                        bytes.Add(currByte);
                        byteOffset = 0;
                        currByte = 0;
                    }
                }
            }

            if (byteOffset > 0)
            {
                bytes.Add(currByte);
            }

            return bytes.ToArray();
        }


        
         * 
         * 
         * 
        public static void SetStruct<T>(byte[] bytes, ref T t) where T : struct
        {
            int offset = 0;
            //object tb = t; //box our object to enable SetValue - don't ask
            PropertyInfo[] fi = t.GetType().GetProperties();
            foreach (var f in fi)
            {
                object[] attrs = f.GetCustomAttributes(typeof(BitfieldAttribute), false);

                if (attrs.Length != 1)
                {
                    throw new Exception("no BitfieldLength-Attribute found");
                }

                int fieldLength = (int)((BitfieldAttribute)attrs[0]).Length;

                ulong part = BinaryPart(bytes, offset, fieldLength);
                offset += fieldLength;

                Type ft = f.PropertyType;

                if (ft.IsValueType && !ft.IsEnum)
                {
                    if (Array.IndexOf(signedTypes, Type.GetTypeCode(f.PropertyType)) != -1)
                    {
                        //item is signed, so we have to fill up the signed bit
                        //first we need to check if it's negative by the bit at the highest possible position
                        if (((part >> (fieldLength - 1)) & 1) == 1)
                        {
                            //it's actually negative, mask it out!
                            //this looks terribly stupid, but it works
                            //shifting left eliminates the 0s on the LSBs
                            part |= ulong.MaxValue << fieldLength;
                        }
                    }


                    switch (Type.GetTypeCode(f.PropertyType))
                    {
                        case TypeCode.Boolean: f.SetValue(t, (bool)(part != 0), null); break;
                        case TypeCode.Char: f.SetValue(t, unchecked((char)part), null); break;
                        case TypeCode.Byte: f.SetValue(t, (byte)part, null); break;
                        case TypeCode.UInt16: f.SetValue(t, unchecked((UInt16)part), null); break;
                        case TypeCode.UInt32: f.SetValue(t, unchecked((UInt32)part), null); break;
                        case TypeCode.UInt64: f.SetValue(t, unchecked((UInt64)part), null); break;
                        case TypeCode.SByte: f.SetValue(t, unchecked((sbyte)part), null); break;
                        case TypeCode.Int16: f.SetValue(t, unchecked((Int16)part), null); break;
                        case TypeCode.Int32: f.SetValue(t, unchecked((Int32)part), null); break;
                        case TypeCode.Int64: f.SetValue(t, unchecked((Int64)part), null); break;
                        //case TypeCode.Decimal:
                        //case TypeCode.Double:
                        //case TypeCode.Single:
                        //case TypeCode.String:
                        default:
                            throw new Exception("phuck! no so value type available.");
                    }
                }
                else if (ft.IsEnum)
                {
                    //this is very very very hacky
                    Int32 meh = unchecked((Int32)part);
                    if (Enum.IsDefined(ft, meh))
                    {
                        f.SetValue(t, meh, null);
                    }
                    else
                        throw new Exception("phuck! could not set enum :(");

                }

                
                
            }
            //t = (T)tb; //unbox it

        }/**/
        

        /// <summary>
        /// slices out /length/ bytes from /offset/ in the byte array and returns them as ulong
        /// </summary>
        /// <param name="bytes"></param>
        /// <param name="offset"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        private static ulong BinaryPart(byte[] bytes, int offset, int length)
        {
            ulong retval = 0;
            int outoffset = 0;

            while (length > 0)
            {
                int bytepos = offset >> 3;
                int bitpos = offset & 0x07;
                int availableBits = 8 - bitpos;

                if (bytes.Length <= bytepos) break;

                byte part = (byte)(bytes[bytepos] >> bitpos);

                byte mask = (byte)((1 << Math.Min(8, length)) - 1);
                part &= mask;

                retval |= (ulong)part << outoffset;

                outoffset += availableBits;
                offset += availableBits;
                length -= availableBits;

            }

            return retval;
        }


    }

}
