﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Security.Permissions;
using System.Windows.Forms;

namespace heohelper
{
    //https://stackoverflow.com/questions/1799370/getting-attributes-of-enums-value
    public static class EnumHelper
    {
        /// <summary>
        /// Gets an attribute on an enum field value
        /// </summary>
        /// <typeparam name="T">The type of the attribute you want to retrieve</typeparam>
        /// <param name="enumVal">The enum value</param>
        /// <returns>The attribute of type T that exists on the enum value</returns>
        /// <example>string desc = myEnumVariable.GetAttributeOfType<DescriptionAttribute>().Description;</example>
        public static T GetAttributeOfType<T>(this Enum enumVal) where T : System.Attribute
        {
            var type = enumVal.GetType();
            var memInfo = type.GetMember(enumVal.ToString());
            var attributes = memInfo[0].GetCustomAttributes(typeof(T), false);
            return (attributes.Length > 0) ? (T)attributes[0] : null;
        }

        public static void foo()
        {
            //typeof(System.Windows.Forms.HorizontalAlignment).GetMembers 

        }

        public static void PopulateCombobox(Type type, ComboBox target, System.Enum selectedValue = null, bool selectFirst = false)
        {
            bool itemselected = false;
            foreach (System.Enum r in Enum.GetValues(type))
            {
                FriendlyNameAttribute fna = EnumHelper.GetAttributeOfType<FriendlyNameAttribute>(r);
                string fn = "";
                if (fna != null)
                    fn = fna.DisplayName;
                else
                    fn = r.ToString();
                Helper.ObjectItem oi = new Helper.ObjectItem(r, fn);
                target.Items.Add(oi);
                if (r.Equals(selectedValue))
                {
                    target.SelectedItem = oi;
                    itemselected = true;
                }
            }
            if (!itemselected && selectFirst)
                target.SelectedIndex = 0;
        }
    }


    [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1813:AvoidUnsealedAttributes")]
    [AttributeUsage(AttributeTargets.All)]
    public class FriendlyNameAttribute : Attribute
    {
        /// <devdoc>
        /// <para>Specifies the default value for the <see cref="System.ComponentModel.DisplayNameAttribute"> , which is an
        ///    empty string (""). This <see langword="static"> field is read-only.</see></see></para>
        /// </devdoc> 
        public static readonly FriendlyNameAttribute Default = new FriendlyNameAttribute();
        private string _friendlyName;

        /// <devdoc>
        ///    <para>[To be supplied.]</para> 
        /// </devdoc>
        public FriendlyNameAttribute() : this(string.Empty)
        {
        }

        /// <devdoc>
        ///    <para>Initializes a new instance of the <see cref="System.ComponentModel.DisplayNameAttribute"> class.</see></para> 
        /// </devdoc> 
        public FriendlyNameAttribute(string friendlyName)
        {
            this._friendlyName = friendlyName;
        }

        /// <devdoc>
        ///    <para>Gets the description stored in this attribute.</para> 
        /// </devdoc>
        public virtual string DisplayName
        {
            get
            {
                return FriendlyNameValue;
            }
        }

        /// <devdoc>
        ///     Read/Write property that directly modifies the string stored 
        ///     in the description attribute. The default implementation
        ///     of the Description property simply returns this value. 
        /// </devdoc> 
        protected string FriendlyNameValue
        {
            get
            {
                return _friendlyName;
            }
            set
            {
                _friendlyName = value;
            }
        }

        public override bool Equals(object obj)
        {
            if (obj == this)
            {
                return true;
            }

            FriendlyNameAttribute other = obj as FriendlyNameAttribute;

            return (other != null) && other.DisplayName == DisplayName;
        }

        public override int GetHashCode()
        {
            return DisplayName.GetHashCode();
        }

        /// <internalonly> 
        /// <devdoc>
        /// </devdoc> 
        public override bool IsDefaultAttribute()
        {
            return (this.Equals(Default));
        }

    }
}
