﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Text.RegularExpressions;

namespace heohelper
{
    public class Helper
    {
        public class SiPrefix
        {
            public string Symbol { get; private set; }
            public string Name { get; private set; }
            public int Exponent { get; private set; }

            public SiPrefix(string symbol, string name, int exponent)
            {
                Symbol = symbol;
                Name = name;
                Exponent = exponent;
            }
        }

        public static readonly List<SiPrefix> SiPrefixes = new List<SiPrefix>()
        {
            new SiPrefix("Y", "Yotta",  24),
            new SiPrefix("Z", "Zetta",  21),
            new SiPrefix("E", "Exa",    18),
            new SiPrefix("P", "Peta",   15),
            new SiPrefix("T", "Tera",   12),
            new SiPrefix("G", "Giga",    9),
            new SiPrefix("M", "Mega",    6),
            new SiPrefix("k", "Kilo",    3),
            new SiPrefix("" , "",        0),
            new SiPrefix("m", "Milli",  -3),
            new SiPrefix("µ", "Mikro",  -6),
            new SiPrefix("n", "Nano",   -9),
            new SiPrefix("p", "Piko",  -12),
            new SiPrefix("f", "Femto", -15),
            new SiPrefix("a", "Atto",  -18),
            new SiPrefix("z", "Zepto", -21),
            new SiPrefix("y", "Yokto", -24)
        };

        public static SiPrefix GetSiPrefix(double value)
        {
            int exp = (int)Math.Round(Math.Log10(value));
            SiPrefix prefix = null;
            for (int i = 0; i < SiPrefixes.Count - 1; i++)
            {
                if(exp < SiPrefixes[i].Exponent && exp >= SiPrefixes[i+1].Exponent)
                {
                    return SiPrefixes[i + 1];
                }
            }
            if(prefix == null)
            {
                if (exp > 0)
                    return SiPrefixes.First();
                else
                    return SiPrefixes.Last();
            }
            return null;
        }

        public static string ByteArrayToHexString(byte[] ba)
        {
            StringBuilder hex = new StringBuilder(ba.Length * 2);
            foreach (byte b in ba)
                hex.AppendFormat("{0:X2} ", b);

            return hex.ToString().Trim();
        }

        public static string ByteArrayToHexString(byte[] ba, string prefix = "", string separator = " ")
        {
            StringBuilder hex = new StringBuilder((ba.Length + prefix.Length + separator.Length) * 2);
            foreach (byte b in ba)
                hex.AppendFormat("{0}{1:X2}{2}", prefix, b, separator);

            if (hex.Length > 0)
                hex.Remove(hex.Length - separator.Length, separator.Length);

            return hex.ToString();
        }

        public static string ByteArrayToBinString(byte[] ba)
        {
            StringBuilder str = new StringBuilder(ba.Length * 2);
            foreach (byte b in ba)
                str.AppendFormat("{0} ", Convert.ToString(b, 2).PadLeft(8, '0'));

            return str.ToString().Trim();
        }

        public static byte[] HexStringToByteArray(string str)
        {
            List<byte> data = new List<byte>();
            Regex re = new Regex("[0-9A-F]{1,2}", RegexOptions.IgnoreCase | RegexOptions.Compiled);
            MatchCollection mc = re.Matches(str);

            foreach(Match m in mc)
            {
                data.Add(Convert.ToByte(m.Value, 16));
            }

            return data.ToArray();
        }


        public struct EnumEntry
        {
            public string DisplayName;
            public string Name;

            public override string ToString()
            {
                return DisplayName;
            }
        }

        public static EnumEntry[] GetEnumEntries(Type type)
        {
            string[] names = Enum.GetNames(type);
            string displayName = "";
            List<EnumEntry> retval = new List<EnumEntry>();

            foreach (string name in names)
            {
                System.Reflection.MemberInfo[] memInfo = type.GetMember(name);
                object[] attributes = memInfo[0].GetCustomAttributes(typeof(DisplayNameExAttribute), false);
                if (attributes.Length >= 1)
                    displayName = ((DisplayNameExAttribute)attributes[0]).DisplayName;
                else
                    displayName = name;

                retval.Add(new EnumEntry() { DisplayName = displayName, Name = name });


            }

            return retval.ToArray();
        }

        public class IntItem
        {
            public int Value;
            public string StrValue;

            public override string ToString()
            {
                return StrValue;
            }

            public IntItem(int value, string strValue)
            {
                Value = value;
                StrValue = strValue;
            }
        }

        public class StringItem
        {
            public string Value;
            public string StrValue;

            public override string ToString()
            {
                return StrValue;
            }

            public StringItem(string value, string strValue)
            {
                Value = value;
                StrValue = strValue;
            }
        }

        public class ObjectItem
        {
            public object Value;
            public string StrValue;

            public override string ToString()
            {
                return StrValue;
            }

            public ObjectItem(object value, string strValue)
            {
                Value = value;
                StrValue = strValue;
            }
        }


        /// <summary>
        /// InvokeIfRequired(this, () =>
        /// {
        /// });
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="action"></param>
        public static void InvokeIfRequired(ISynchronizeInvoke obj, MethodInvoker action)
        {
            if (obj.InvokeRequired)
            {
                var args = new object[0];
                try
                {
                    obj.Invoke(action, args);
                }
                catch { }
            }
            else
            {
                action();
            }
        }

    }
}
